def p2p_port(base_port):
    return base_port


def rpc_port(base_port):
    return base_port + 7


def grpc_port(base_port):
    return base_port + 3


def api_port(base_port):
    return base_port + 4


def pprof_port(base_port):
    return base_port + 5


def grpc_port_tx_only(base_port):
    return base_port + 6



import enum
import hashlib
import json
import tempfile
import threading
import time

import bech32
from dateutil.parser import isoparse

from .app import CHAIN
from .ledger import ZEMU_BUTTON_PORT, ZEMU_HOST, LedgerButton
from .utils import build_cli_args_safe, format_doc_string, interact


class ModuleAccount(enum.Enum):
    FeeCollector = "fee_collector"
    Mint = "mint"
    Gov = "gov"
    Distribution = "distribution"
    BondedPool = "bonded_tokens_pool"
    NotBondedPool = "not_bonded_tokens_pool"
    IBCTransfer = "transfer"


@format_doc_string(
    options=",".join(v.value for v in ModuleAccount.__members__.values())
)
def module_address(name):
    """
    get address of module accounts

    :param name: name of module account, values: {options}
    """
    data = hashlib.sha256(ModuleAccount(name).value.encode()).digest()[:20]
    return bech32.bech32_encode("cro", bech32.convertbits(data, 8, 5))


class ChainCommand:
    def __init__(self, cmd=None):
        self.cmd = cmd or CHAIN

    def __call__(self, cmd, *args, stdin=None, **kwargs):
        "execute chain-maind"
        args = " ".join(build_cli_args_safe(cmd, *args, **kwargs))
        return interact(f"{self.cmd} {args}", input=stdin)


class CosmosCLI:
    "the apis to interact with wallet and blockchain"

    def __init__(
        self,
        data_dir,
        node_rpc,
        chain_id=None,
        cmd=None,
        zemu_address=ZEMU_HOST,
        zemu_button_port=ZEMU_BUTTON_PORT,
    ):
        self.data_dir = data_dir
        if chain_id is None:
            self._genesis = json.load(open(self.data_dir / "config" / "genesis.json"))
            self.chain_id = self._genesis["chain_id"]
        else:
            self.chain_id = chain_id
        self.node_rpc = node_rpc
        self.raw = ChainCommand(cmd)
        self.leger_button = LedgerButton(zemu_address, zemu_button_port)
        self.output = None
        self.error = None

    def node_id(self):
        "get tendermint node id"
        output = self.raw("tendermint", "show-node-id", home=self.data_dir)
        return output.decode().strip()

    def delete_account(self, name):
        "delete wallet account in node's keyring"
        return self.raw(
            "keys",
            "delete",
            name,
            "-y",
            "--force",
            home=self.data_dir,
            output="json",
            keyring_backend="test",
        )

    def create_account(self, name, mnemonic=None):
        "create new keypair in node's keyring"
        if mnemonic is None:
            output = self.raw(
                "keys",
                "add",
                name,
                home=self.data_dir,
                output="json",
                keyring_backend="test",
            )
        else:
            output = self.raw(
                "keys",
                "add",
                name,
                "--recover",
                home=self.data_dir,
                output="json",
                keyring_backend="test",
                stdin=mnemonic.encode() + b"\n",
            )
        return json.loads(output)

    def create_account_ledger(self, name):
        "create new ledger keypair"

        def send_request():
            try:
                self.output = self.raw(
                    "keys",
                    "add",
                    name,
                    "--ledger",
                    home=self.data_dir,
                    output="json",
                    keyring_backend="test",
                )
            except Exception as e:
                self.error = e

        t = threading.Thread(target=send_request)
        t.start()
        time.sleep(3)
        for _ in range(0, 3):
            self.leger_button.press_right()
            time.sleep(0.2)
        self.leger_button.press_both()
        t.join()
        if self.error:
            raise self.error
        return json.loads(self.output)

    def init(self, moniker):
        "the node's config is already added"
        return self.raw(
            "init",
            moniker,
            chain_id=self.chain_id,
            home=self.data_dir,
        )

    def validate_genesis(self):
        return self.raw("validate-genesis", home=self.data_dir)

    def add_genesis_account(self, addr, coins, **kwargs):
        return self.raw(
            "add-genesis-account",
            addr,
            coins,
            home=self.data_dir,
            output="json",
            **kwargs,
        )

    def gentx(self, name, coins, min_self_delegation=1, pubkey=None):
        return self.raw(
            "gentx",
            name,
            coins,
            min_self_delegation=str(min_self_delegation),
            home=self.data_dir,
            chain_id=self.chain_id,
            keyring_backend="test",
            pubkey=pubkey,
        )

    def collect_gentxs(self, gentx_dir):
        return self.raw("collect-gentxs", gentx_dir, home=self.data_dir)

    def status(self):
        return json.loads(self.raw("status", node=self.node_rpc))

    def block_height(self):
        return int(self.status()["SyncInfo"]["latest_block_height"])

    def block_time(self):
        return isoparse(self.status()["SyncInfo"]["latest_block_time"])

    def balance(self, addr):
        coin = json.loads(
            self.raw(
                "query", "bank", "balances", addr, output="json", node=self.node_rpc
            )
        )["balances"]
        if len(coin) == 0:
            return 0
        coin = coin[0]
        return int(coin["amount"])

    def query_all_txs(self, addr):
        txs = self.raw(
            "query",
            "txs-all",
            addr,
            home=self.data_dir,
            keyring_backend="test",
            chain_id=self.chain_id,
            node=self.node_rpc,
        )
        return json.loads(txs)

    def distribution_commission(self, addr):
        coin = json.loads(
            self.raw(
                "query",
                "distribution",
                "commission",
                addr,
                output="json",
                node=self.node_rpc,
            )
        )["commission"][0]
        return float(coin["amount"])

    def distribution_community(self):
        coin = json.loads(
            self.raw(
                "query",
                "distribution",
                "community-pool",
                output="json",
                node=self.node_rpc,
            )
        )["pool"][0]
        return float(coin["amount"])

    def distribution_reward(self, delegator_addr):
        coin = json.loads(
            self.raw(
                "query",
                "distribution",
                "rewards",
                delegator_addr,
                output="json",
                node=self.node_rpc,
            )
        )["total"][0]
        return float(coin["amount"])

    def address(self, name, bech="acc"):
        output = self.raw(
            "keys",
            "show",
            name,
            "-a",
            home=self.data_dir,
            keyring_backend="test",
            bech=bech,
        )
        return output.strip().decode()

    def account(self, addr):
        return json.loads(
            self.raw(
                "query", "auth", "account", addr, output="json", node=self.node_rpc
            )
        )

    def supply(self, supply_type):
        return json.loads(
            self.raw("query", "supply", supply_type, output="json", node=self.node_rpc)
        )

    def validator(self, addr):
        return json.loads(
            self.raw(
                "query",
                "staking",
                "validator",
                addr,
                output="json",
                node=self.node_rpc,
            )
        )

    def validators(self):
        return json.loads(
            self.raw(
                "query", "staking", "validators", output="json", node=self.node_rpc
            )
        )["validators"]

    def staking_params(self):
        return json.loads(
            self.raw("query", "staking", "params", output="json", node=self.node_rpc)
        )

    def staking_pool(self, bonded=True):
        return int(
            json.loads(
                self.raw("query", "staking", "pool", output="json", node=self.node_rpc)
            )["bonded_tokens" if bonded else "not_bonded_tokens"]
        )

    def transfer(self, from_, to, coins, generate_only=False, fees=None):
        return json.loads(
            self.raw(
                "tx",
                "bank",
                "send",
                from_,
                to,
                coins,
                "-y",
                "--generate-only" if generate_only else None,
                home=self.data_dir,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
                fees=fees,
            )
        )

    def transfer_from_ledger(self, from_, to, coins, generate_only=False, fees=None):
        def send_request():
            try:
                self.output = self.raw(
                    "tx",
                    "bank",
                    "send",
                    from_,
                    to,
                    coins,
                    "-y",
                    "--generate-only" if generate_only else "",
                    "--ledger",
                    home=self.data_dir,
                    keyring_backend="test",
                    chain_id=self.chain_id,
                    node=self.node_rpc,
                    fees=fees,
                    sign_mode="amino-json",
                )
            except Exception as e:
                self.error = e

        t = threading.Thread(target=send_request)
        t.start()
        time.sleep(3)
        for _ in range(0, 11):
            self.leger_button.press_right()
            time.sleep(0.4)
        self.leger_button.press_both()
        t.join()
        if self.error:
            raise self.error
        return json.loads(self.output)

    def get_delegated_amount(self, which_addr):
        return json.loads(
            self.raw(
                "query",
                "staking",
                "delegations",
                which_addr,
                home=self.data_dir,
                chain_id=self.chain_id,
                node=self.node_rpc,
                output="json",
            )
        )

    def delegate_amount(self, to_addr, amount, from_addr):
        return json.loads(
            self.raw(
                "tx",
                "staking",
                "delegate",
                to_addr,
                amount,
                "-y",
                home=self.data_dir,
                from_=from_addr,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    # to_addr: croclcl1...  , from_addr: cro1...
    def unbond_amount(self, to_addr, amount, from_addr):
        return json.loads(
            self.raw(
                "tx",
                "staking",
                "unbond",
                to_addr,
                amount,
                "-y",
                home=self.data_dir,
                from_=from_addr,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    # to_validator_addr: crocncl1...  ,  from_from_validator_addraddr: crocl1...
    def redelegate_amount(
        self, to_validator_addr, from_validator_addr, amount, from_addr
    ):
        return json.loads(
            self.raw(
                "tx",
                "staking",
                "redelegate",
                from_validator_addr,
                to_validator_addr,
                amount,
                "-y",
                home=self.data_dir,
                from_=from_addr,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    # from_delegator can be account name or address
    def withdraw_all_rewards(self, from_delegator):
        return json.loads(
            self.raw(
                "tx",
                "distribution",
                "withdraw-all-rewards",
                "-y",
                from_=from_delegator,
                home=self.data_dir,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    def make_multisig(self, name, signer1, signer2):
        self.raw(
            "keys",
            "add",
            name,
            multisig=f"{signer1},{signer2}",
            multisig_threshold="2",
            home=self.data_dir,
            keyring_backend="test",
            output="json",
        )

    def sign_multisig_tx(self, tx_file, multi_addr, signer_name):
        return json.loads(
            self.raw(
                "tx",
                "sign",
                tx_file,
                from_=signer_name,
                multisig=multi_addr,
                home=self.data_dir,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    def sign_batch_multisig_tx(
        self, tx_file, multi_addr, signer_name, account_number, sequence_number
    ):
        r = self.raw(
            "tx",
            "sign-batch",
            "--offline",
            tx_file,
            account_number=account_number,
            sequence=sequence_number,
            from_=signer_name,
            multisig=multi_addr,
            home=self.data_dir,
            keyring_backend="test",
            chain_id=self.chain_id,
            node=self.node_rpc,
        )
        return r.decode("utf-8")

    def encode_signed_tx(self, signed_tx):
        return self.raw(
            "tx",
            "encode",
            signed_tx,
        )

    def sign_single_tx(self, tx_file, signer_name):
        return json.loads(
            self.raw(
                "tx",
                "sign",
                tx_file,
                from_=signer_name,
                home=self.data_dir,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    def combine_multisig_tx(self, tx_file, multi_name, signer1_file, signer2_file):
        return json.loads(
            self.raw(
                "tx",
                "multisign",
                tx_file,
                multi_name,
                signer1_file,
                signer2_file,
                home=self.data_dir,
                keyring_backend="test",
                chain_id=self.chain_id,
                node=self.node_rpc,
            )
        )

    def combine_batch_multisig_tx(
        self, tx_file, multi_name, signer1_file, signer2_file
    ):
        r = self.raw(
            "tx",
            "multisign-batch",
            tx_file,
            multi_name,
            signer1_file,
            signer2_file,
            home=self.data_dir,
            keyring_backend="test",
            chain_id=self.chain_id,
            node=self.node_rpc,
        )
        return r.decode("utf-8")

    def broadcast_tx(self, tx_file):
        return json.loads(self.raw("tx", "broadcast", tx_file, node=self.node_rpc))

    def unjail(self, addr):
        return json.loads(
            self.raw(
                "tx",
                "slashing",
                "unjail",
                "-y",
                from_=addr,
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
            )
        )

    def create_validator(
        self,
        amount,
        moniker=None,
        commission_max_change_rate="0.01",
        commission_rate="0.1",
        commission_max_rate="0.2",
        min_self_delegation="1",
        identity="",
        website="",
        security_contact="",
        details="",
    ):
        """MsgCreateValidator
        create the node with create_node before call this"""
        pubkey = (
            self.raw(
                "tendermint",
                "show-validator",
                home=self.data_dir,
            )
            .strip()
            .decode()
        )
        return json.loads(
            self.raw(
                "tx",
                "staking",
                "create-validator",
                "-y",
                from_=self.address("validator"),
                amount=amount,
                pubkey=pubkey,
                min_self_delegation=min_self_delegation,
                # commision
                commission_rate=commission_rate,
                commission_max_rate=commission_max_rate,
                commission_max_change_rate=commission_max_change_rate,
                # description
                moniker=moniker,
                identity=identity,
                website=website,
                security_contact=security_contact,
                details=details,
                # basic
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
            )
        )

    def edit_validator(
        self,
        commission_rate=None,
        moniker=None,
        identity=None,
        website=None,
        security_contact=None,
        details=None,
    ):
        """MsgEditValidator"""
        options = dict(
            commission_rate=commission_rate,
            # description
            moniker=moniker,
            identity=identity,
            website=website,
            security_contact=security_contact,
            details=details,
        )
        return json.loads(
            self.raw(
                "tx",
                "staking",
                "edit-validator",
                "-y",
                from_=self.address("validator"),
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
                **{k: v for k, v in options.items() if v is not None},
            )
        )

    def gov_propose(self, proposer, kind, proposal):
        if kind == "software-upgrade":
            return json.loads(
                self.raw(
                    "tx",
                    "gov",
                    "submit-proposal",
                    kind,
                    proposal["name"],
                    "-y",
                    from_=proposer,
                    # content
                    title=proposal.get("title"),
                    description=proposal.get("description"),
                    upgrade_height=proposal.get("upgrade-height"),
                    upgrade_time=proposal.get("upgrade-time"),
                    upgrade_info=proposal.get("upgrade-info"),
                    deposit=proposal.get("deposit"),
                    # basic
                    home=self.data_dir,
                    node=self.node_rpc,
                    keyring_backend="test",
                    chain_id=self.chain_id,
                )
            )
        elif kind == "cancel-software-upgrade":
            return json.loads(
                self.raw(
                    "tx",
                    "gov",
                    "submit-proposal",
                    kind,
                    "-y",
                    from_=proposer,
                    # content
                    title=proposal.get("title"),
                    description=proposal.get("description"),
                    deposit=proposal.get("deposit"),
                    # basic
                    home=self.data_dir,
                    node=self.node_rpc,
                    keyring_backend="test",
                    chain_id=self.chain_id,
                )
            )
        else:
            with tempfile.NamedTemporaryFile("w") as fp:
                json.dump(proposal, fp)
                fp.flush()
                return json.loads(
                    self.raw(
                        "tx",
                        "gov",
                        "submit-proposal",
                        kind,
                        fp.name,
                        "-y",
                        from_=proposer,
                        # basic
                        home=self.data_dir,
                        node=self.node_rpc,
                        keyring_backend="test",
                        chain_id=self.chain_id,
                    )
                )

    def gov_vote(self, voter, proposal_id, option):
        print(voter)
        print(proposal_id)
        print(option)
        return json.loads(
            self.raw(
                "tx",
                "gov",
                "vote",
                proposal_id,
                option,
                "-y",
                from_=voter,
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
            )
        )

    def gov_deposit(self, depositor, proposal_id, amount):
        return json.loads(
            self.raw(
                "tx",
                "gov",
                "deposit",
                proposal_id,
                amount,
                "-y",
                from_=depositor,
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
            )
        )

    def query_proposals(self, depositor=None, limit=None, status=None, voter=None):
        return json.loads(
            self.raw(
                "query",
                "gov",
                "proposals",
                depositor=depositor,
                count_total=limit,
                status=status,
                voter=voter,
                output="json",
                node=self.node_rpc,
            )
        )

    def query_proposal(self, proposal_id):
        return json.loads(
            self.raw(
                "query",
                "gov",
                "proposal",
                proposal_id,
                output="json",
                node=self.node_rpc,
            )
        )

    def query_tally(self, proposal_id):
        return json.loads(
            self.raw(
                "query",
                "gov",
                "tally",
                proposal_id,
                output="json",
                node=self.node_rpc,
            )
        )

    def ibc_transfer(
        self,
        from_,
        to,
        amount,
        channel,  # src channel
        target_version,  # chain version number of target chain
        i=0,
    ):
        return json.loads(
            self.raw(
                "tx",
                "ibc-transfer",
                "transfer",
                "transfer",  # src port
                channel,
                to,
                amount,
                "-y",
                # FIXME https://github.com/cosmos/cosmos-sdk/issues/8059
                "--absolute-timeouts",
                from_=from_,
                home=self.data_dir,
                node=self.node_rpc,
                keyring_backend="test",
                chain_id=self.chain_id,
                packet_timeout_height=f"{target_version}-10000000000",
                packet_timeout_timestamp=0,
            )
        )

    def export(self):
        return self.raw("export", home=self.data_dir)

    def unsaferesetall(self):
        return self.raw("unsafe-reset-all")



import base64

import cosmos.bank.v1beta1.tx_pb2
import cosmos.bank.v1beta1.tx_pb2_grpc
import cosmos.crypto.ed25519.keys_pb2
import cosmos.staking.v1beta1.query_pb2
import cosmos.staking.v1beta1.query_pb2_grpc
import grpc
import tendermint.rpc.grpc.types_pb2_grpc


# for query only
class GrpcUtil:
    def __init__(self, ip_port):
        self.ip_port = ip_port

    def get_validators(self):
        channel = grpc.insecure_channel(self.ip_port)
        stub = cosmos.staking.v1beta1.query_pb2_grpc.QueryStub(channel)
        response = stub.Validators(
            cosmos.staking.v1beta1.query_pb2.QueryValidatorsRequest()
        )
        return response


# for tx broadcast only
class GrpcUtilTxBroadcast:
    def __init__(self, ip_port):
        self.ip_port = ip_port

    def send_tx_in_base64(self, tx_base64):
        tx_raw_bytes = base64.b64decode(tx_base64)
        channel = grpc.insecure_channel(self.ip_port)
        stub = tendermint.rpc.grpc.types_pb2_grpc.BroadcastAPIStub(channel)
        request = tendermint.rpc.grpc.types_pb2.RequestBroadcastTx()
        request.tx = tx_raw_bytes
        response = stub.BroadcastTx(request)
        return response



import base64
import configparser
import datetime
import hashlib
import json
import os
import re
import subprocess
import sys
import threading
import time
from pathlib import Path
from typing import List

import durations
import jsonmerge
import multitail2
import tomlkit
import yaml
from dateutil.parser import isoparse
from supervisor import xmlrpc
from supervisor.compat import xmlrpclib

from . import ports
from .app import CHAIN, IMAGE, SUPERVISOR_CONFIG_FILE
from .cosmoscli import ChainCommand, CosmosCLI, ModuleAccount, module_address
from .ledger import ZEMU_BUTTON_PORT, ZEMU_HOST
from .utils import format_doc_string, interact, write_ini

COMMON_PROG_OPTIONS = {
    # redirect to supervisord's stdout, easier to collect all logs
    "autostart": "true",
    "autorestart": "true",
    "redirect_stderr": "true",
    "startsecs": "3",
}


def home_dir(data_dir, i):
    return data_dir / f"node{i}"


class ClusterCLI:
    "the apis to interact with wallet and blockchain prepared with Cluster"

    def __init__(
        self,
        data,
        chain_id="chainmaind",
        cmd=CHAIN,
        zemu_address=ZEMU_HOST,
        zemu_button_port=ZEMU_BUTTON_PORT,
    ):
        self.data_root = data
        self.cmd = cmd
        self.zemu_address = zemu_address
        self.zemu_button_port = zemu_button_port
        self.chain_id = chain_id
        self.data_dir = data / self.chain_id
        self.config = json.load((self.data_dir / "config.json").open())

        self._supervisorctl = None
        self.output = None
        self.error = None

    def cosmos_cli(self, i=0):
        return CosmosCLI(
            self.home(i),
            self.node_rpc(i),
            chain_id=self.chain_id,
            cmd=self.cmd,
            zemu_address=self.zemu_address,
            zemu_button_port=self.zemu_button_port,
        )

    @property
    def supervisor(self):
        "http://supervisord.org/api.html"
        # copy from:
        # https://github.com/Supervisor/supervisor/blob/76df237032f7d9fbe80a0adce3829c8b916d5b58/supervisor/options.py#L1718
        if self._supervisorctl is None:
            self._supervisorctl = xmlrpclib.ServerProxy(
                # dumbass ServerProxy won't allow us to pass in a non-HTTP url,
                # so we fake the url we pass into it and
                # always use the transport's
                # 'serverurl' to figure out what to attach to
                "http://127.0.0.1",
                transport=xmlrpc.SupervisorTransport(
                    serverurl=f"unix://{self.data_root}/supervisor.sock"
                ),
            )
        return self._supervisorctl.supervisor

    def reload_supervisor(self):
        subprocess.run(
            [
                sys.executable,
                "-msupervisor.supervisorctl",
                "-c",
                self.data_root / SUPERVISOR_CONFIG_FILE,
                "update",
            ],
            check=True,
        )

    def nodes_len(self):
        "find how many 'node{i}' sub-directories"
        return len(
            [p for p in self.data_dir.iterdir() if re.match(r"^node\d+$", p.name)]
        )

    def copy_validator_key(self, from_node=1, to_node=2):
        "Copy the validtor file in from_node to to_node"
        from_key_file = "{}/node{}/config/priv_validator_key.json".format(
            self.data_dir, from_node
        )
        to_key_file = "{}/node{}/config/priv_validator_key.json".format(
            self.data_dir, to_node
        )
        with open(from_key_file, "r") as f:
            key = f.read()
        with open(to_key_file, "w") as f:
            f.write(key)

    def update_genesis(self, i, genesis_data):
        home = self.home(i)
        genesis_file = home / "config/genesis.json"
        with open(genesis_file, "w") as f:
            f.write(json.dumps(genesis_data, indent=4))

    def stop_node(self, i=0):
        subprocess.run(
            [
                sys.executable,
                "-msupervisor.supervisorctl",
                "-c",
                self.data_root / SUPERVISOR_CONFIG_FILE,
                "stop",
                "{}-node{}".format(self.chain_id, i),
            ]
        )

    def stop_relayer(self, path):
        subprocess.run(
            [
                sys.executable,
                "-msupervisor.supervisorctl",
                "-c",
                self.data_root / SUPERVISOR_CONFIG_FILE,
                "stop",
                "program:relayer-{}".format(path),
            ]
        )

    def restart_relayer(self, path):
        subprocess.run(
            [
                sys.executable,
                "-msupervisor.supervisorctl",
                "-c",
                self.data_root / SUPERVISOR_CONFIG_FILE,
                "restart",
                "program:relayer-{}".format(path),
            ]
        )

    def start_node(self, i):
        subprocess.run(
            [
                sys.executable,
                "-msupervisor.supervisorctl",
                "-c",
                self.data_root / SUPERVISOR_CONFIG_FILE,
                "start",
                "{}-node{}".format(self.chain_id, i),
            ]
        )

    def create_node(
        self,
        base_port=None,
        moniker=None,
        hostname="localhost",
        statesync=False,
        mnemonic=None,
    ):
        """create new node in the data directory,
        process information is written into supervisor config
        start it manually with supervisor commands

        :return: new node index and config
        """
        i = self.nodes_len()

        # default configs
        if base_port is None:
            # use the node0's base_port + i * 10 as default base port for new ndoe
            base_port = self.config["validators"][0]["base_port"] + i * 10
        if moniker is None:
            moniker = f"node{i}"

        # add config
        assert len(self.config["validators"]) == i
        self.config["validators"].append(
            {
                "base_port": base_port,
                "hostname": hostname,
                "moniker": moniker,
            }
        )
        (self.data_dir / "config.json").write_text(json.dumps(self.config))

        # init home directory
        self.init(i)
        home = self.home(i)
        (home / "config/genesis.json").unlink()
        (home / "config/genesis.json").symlink_to("../../genesis.json")
        # use p2p peers from node0's config
        node0 = tomlkit.parse((self.data_dir / "node0/config/config.toml").read_text())

        def custom_edit_tm(doc):
            if statesync:
                info = self.status()["SyncInfo"]
                doc["statesync"].update(
                    {
                        "enable": True,
                        "rpc_servers": ",".join(self.node_rpc(i) for i in range(2)),
                        "trust_height": int(info["earliest_block_height"]),
                        "trust_hash": info["earliest_block_hash"],
                        "temp_dir": str(self.data_dir),
                        "discovery_time": "5s",
                    }
                )

        edit_tm_cfg(
            home / "config/config.toml",
            base_port,
            node0["p2p"]["persistent_peers"],
            custom_edit=custom_edit_tm,
        )
        edit_app_cfg(home / "config/app.toml", base_port)

        # create validator account
        self.create_account("validator", i, mnemonic)

        # add process config into supervisor
        path = self.data_dir / SUPERVISOR_CONFIG_FILE
        ini = configparser.RawConfigParser()
        ini.read_file(path.open())
        chain_id = self.config["chain_id"]
        prgname = f"{chain_id}-node{i}"
        section = f"program:{prgname}"
        ini.add_section(section)
        ini[section].update(
            dict(
                COMMON_PROG_OPTIONS,
                command=f"{self.cmd} start --home %(here)s/node{i}",
                autostart="false",
                stdout_logfile=f"%(here)s/node{i}.log",
            )
        )
        with path.open("w") as fp:
            ini.write(fp)
        self.reload_supervisor()
        return i

    def home(self, i):
        "home directory of i-th node"
        return home_dir(self.data_dir, i)

    def base_port(self, i):
        return self.config["validators"][i]["base_port"]

    def node_rpc(self, i):
        "rpc url of i-th node"
        return "tcp://127.0.0.1:%d" % ports.rpc_port(self.base_port(i))

    # for query
    def ipport_grpc(self, i):
        "grpc url of i-th node"
        return "127.0.0.1:%d" % ports.grpc_port(self.base_port(i))

    # tx broadcast only
    def ipport_grpc_tx(self, i):
        "grpc url of i-th node"
        return "127.0.0.1:%d" % ports.grpc_port_tx_only(self.base_port(i))

    def node_id(self, i):
        "get i-th node's tendermint node id"
        return self.cosmos_cli(i).node_id()

    def delete_account(self, name, i=0):
        "delete account in i-th node's keyring"
        return self.cosmos_cli(i).delete_account(name)

    def create_account(self, name, i=0, mnemonic=None):
        "create new keypair in i-th node's keyring"
        return self.cosmos_cli(i).create_account(name, mnemonic)

    def create_account_ledger(self, name, i=0):
        "create new ledger keypair"
        return self.cosmos_cli(i).create_account_ledger(name)

    def init(self, i):
        "the i-th node's config is already added"
        return self.cosmos_cli(i).init(self.config["validators"][i]["moniker"])

    def export(self, i=0):
        return self.cosmos_cli(i).export()

    def validate_genesis(self, i=0):
        return self.cosmos_cli(i).validate_genesis()

    def add_genesis_account(self, addr, coins, i=0, **kwargs):
        return self.cosmos_cli(i).add_genesis_account(addr, coins, **kwargs)

    def gentx(self, name, coins, i=0, min_self_delegation=1, pubkey=None):
        return self.cosmos_cli(i).gentx(name, coins, min_self_delegation, pubkey)

    def collect_gentxs(self, gentx_dir, i=0):
        return self.cosmos_cli(i).collect_gentxs(gentx_dir)

    def status(self, i=0):
        return self.cosmos_cli(i).status()

    def block_height(self, i=0):
        return self.cosmos_cli(i).block_height()

    def block_time(self, i=0):
        return self.cosmos_cli(i).block_time()

    def balance(self, addr, i=0):
        return self.cosmos_cli(i).balance(addr)

    def query_all_txs(self, addr, i=0):
        return self.cosmos_cli(i).query_all_txs(addr)

    def distribution_commission(self, addr, i=0):
        return self.cosmos_cli(i).distribution_commission(addr)

    def distribution_community(self, i=0):
        return self.cosmos_cli(i).distribution_community()

    def distribution_reward(self, delegator_addr, i=0):
        return self.cosmos_cli(i).distribution_reward(delegator_addr)

    def address(self, name, i=0, bech="acc"):
        return self.cosmos_cli(i).address(name, bech)

    @format_doc_string(
        options=",".join(v.value for v in ModuleAccount.__members__.values())
    )
    def module_address(self, name):
        """
        get address of module accounts

        :param name: name of module account, values: {options}
        """
        return module_address(name)

    def account(self, addr, i=0):
        return self.cosmos_cli(i).account(addr)

    def supply(self, supply_type, i=0):
        return self.cosmos_cli(i).supply(supply_type)

    def validator(self, addr, i=0):
        return self.cosmos_cli(i).validator(addr)

    def validators(self, i=0):
        return self.cosmos_cli(i).validators()

    def staking_params(self, i=0):
        return self.cosmos_cli(i).staking_params()

    def staking_pool(self, bonded=True, i=0):
        return self.cosmos_cli(i).staking_pool(bonded)

    def transfer_offline(self, from_, to, coins, sequence, i=0, fees=None):
        return self.cosmos_cli(i).transfer_offline(from_, to, coins, sequence, fees)

    def transfer(self, from_, to, coins, i=0, generate_only=False, fees=None):
        return self.cosmos_cli(i).transfer(from_, to, coins, generate_only, fees)

    def transfer_from_ledger(
        self, from_, to, coins, i=0, generate_only=False, fees=None
    ):
        return self.cosmos_cli(i).transfer_from_ledger(
            from_,
            to,
            coins,
            generate_only,
            fees,
        )

    def get_delegated_amount(self, which_addr, i=0):
        return self.cosmos_cli(i).get_delegated_amount(which_addr)

    def delegate_amount(self, to_addr, amount, from_addr, i=0):
        return self.cosmos_cli(i).delegate_amount(to_addr, amount, from_addr)

    # to_addr: croclcl1...  , from_addr: cro1...
    def unbond_amount(self, to_addr, amount, from_addr, i=0):
        return self.cosmos_cli(i).unbond_amount(to_addr, amount, from_addr)

    # to_validator_addr: crocncl1...  ,  from_from_validator_addraddr: crocl1...
    def redelegate_amount(
        self, to_validator_addr, from_validator_addr, amount, from_addr, i=0
    ):
        return self.cosmos_cli(i).redelegate_amount(
            to_validator_addr,
            from_validator_addr,
            amount,
            from_addr,
        )

    def withdraw_all_rewards(self, from_delegator, i=0):
        return self.cosmos_cli(i).withdraw_all_rewards(from_delegator)

    def make_multisig(self, name, signer1, signer2, i=0):
        return self.cosmos_cli(i).make_multisig(name, signer1, signer2)

    def sign_multisig_tx(self, tx_file, multi_addr, signer_name, i=0):
        return self.cosmos_cli(i).sign_multisig_tx(tx_file, multi_addr, signer_name)

    def sign_batch_multisig_tx(
        self, tx_file, multi_addr, signer_name, account_num, sequence, i=0
    ):
        return self.cosmos_cli(i).sign_batch_multisig_tx(
            tx_file, multi_addr, signer_name, account_num, sequence
        )

    def encode_signed_tx(self, signed_tx, i=0):
        return self.cosmos_cli(i).encode_signed_tx(signed_tx)

    def sign_single_tx(self, tx_file, signer_name, i=0):
        return self.cosmos_cli(i).sign_single_tx(tx_file, signer_name)

    def combine_multisig_tx(self, tx_file, multi_name, signer1_file, signer2_file, i=0):
        return self.cosmos_cli(i).combine_multisig_tx(
            tx_file,
            multi_name,
            signer1_file,
            signer2_file,
        )

    def combine_batch_multisig_tx(
        self, tx_file, multi_name, signer1_file, signer2_file, i=0
    ):
        return self.cosmos_cli(i).combine_batch_multisig_tx(
            tx_file,
            multi_name,
            signer1_file,
            signer2_file,
        )

    def broadcast_tx(self, tx_file, i=0):
        return self.cosmos_cli(i).broadcast_tx(tx_file)

    def unjail(self, addr, i=0):
        return self.cosmos_cli(i).unjail(addr)

    def create_validator(
        self,
        amount,
        i,
        moniker=None,
        commission_max_change_rate="0.01",
        commission_rate="0.1",
        commission_max_rate="0.2",
        min_self_delegation="1",
        identity="",
        website="",
        security_contact="",
        details="",
    ):
        """MsgCreateValidator
        create the node with create_node before call this"""
        return self.cosmos_cli(i).create_validator(
            amount,
            moniker or self.config["validators"][i]["moniker"],
            commission_max_change_rate,
            commission_rate,
            commission_max_rate,
            min_self_delegation,
            identity,
            website,
            security_contact,
            details,
        )

    def edit_validator(
        self,
        i,
        commission_rate=None,
        moniker=None,
        identity=None,
        website=None,
        security_contact=None,
        details=None,
    ):
        """MsgEditValidator"""
        return self.cosmos_cli(i).edit_validator(
            commission_rate,
            moniker,
            identity,
            website,
            security_contact,
            details,
        )

    def gov_propose(self, proposer, kind, proposal, i=0):
        return self.cosmos_cli(i).gov_propose(proposer, kind, proposal)

    def gov_vote(self, voter, proposal_id, option, i=0):
        return self.cosmos_cli(i).gov_vote(voter, proposal_id, option)

    def gov_deposit(self, depositor, proposal_id, amount, i=0):
        return self.cosmos_cli(i).gov_deposit(depositor, proposal_id, amount)

    def query_proposals(self, depositor=None, limit=None, status=None, voter=None, i=0):
        return self.cosmos_cli(i).query_proposals(depositor, limit, status, voter)

    def query_proposal(self, proposal_id, i=0):
        return self.cosmos_cli(i).query_proposal(proposal_id)

    def query_tally(self, proposal_id, i=0):
        return self.cosmos_cli(i).query_tally(proposal_id)

    def ibc_transfer(
        self,
        from_,
        to,
        amount,
        channel,  # src channel
        target_version,  # chain version number of target chain
        i=0,
    ):
        return self.cosmos_cli(i).ibc_transfer(
            from_,
            to,
            amount,
            channel,
            target_version,
        )


def start_cluster(data_dir):
    cmd = [
        sys.executable,
        "-msupervisor.supervisord",
        "-c",
        data_dir / SUPERVISOR_CONFIG_FILE,
    ]
    return subprocess.Popen(cmd, env=dict(os.environ, PYTHONPATH=":".join(sys.path)))


class TailLogsThread(threading.Thread):
    def __init__(self, base_dir, pats: List[str]):
        self.base_dir = base_dir
        self.tailer = multitail2.MultiTail([str(base_dir / pat) for pat in pats])
        self._stop_event = threading.Event()
        super().__init__()

    def run(self):
        while not self.stopped:
            for (path, _), s in self.tailer.poll():
                print(Path(path).relative_to(self.base_dir), s)

            # TODO Replace this with FAM/inotify for watching filesystem events.
            time.sleep(0.5)

    def stop(self):
        self._stop_event.set()

    @property
    def stopped(self):
        return self._stop_event.is_set()


def start_tail_logs_thread(data_dir):
    t = TailLogsThread(data_dir, ["*/node*.log", "relayer-*.log"])
    t.start()
    return t


def process_config(config, base_port):
    """
    fill default values in config
    """
    for i, val in enumerate(config["validators"]):
        if "moniker" not in val:
            val["moniker"] = f"node{i}"
        if "base_port" not in val:
            val["base_port"] = base_port + i * 10
        if "hostname" not in val:
            val["hostname"] = "localhost"


def init_devnet(
    data_dir,
    config,
    base_port,
    image=IMAGE,
    cmd=None,
    gen_compose_file=False,
):
    """
    init data directory
    """

    def create_account(cli, account, use_ledger=False):
        if use_ledger:
            acct = cli.create_account_ledger(account["name"])
        else:
            acct = cli.create_account(account["name"])
        vesting = account.get("vesting")
        if not vesting:
            cli.add_genesis_account(acct["address"], account["coins"])
        else:
            genesis_time = isoparse(genesis["genesis_time"])
            end_time = genesis_time + datetime.timedelta(
                seconds=durations.Duration(vesting).to_seconds()
            )
            vend = int(end_time.timestamp())
            cli.add_genesis_account(
                acct["address"],
                account["coins"],
                vesting_amount=account["coins"],
                vesting_end_time=vend,
            )
        return acct

    process_config(config, base_port)

    (data_dir / "config.json").write_text(json.dumps(config))

    cmd = config.get("cmd") or cmd or CHAIN

    # init home directories
    for i, val in enumerate(config["validators"]):
        ChainCommand(cmd)(
            "init",
            val["moniker"],
            chain_id=config["chain_id"],
            home=home_dir(data_dir, i),
        )
        if "consensus_key" in val:
            # restore consensus private key
            with (home_dir(data_dir, i) / "config/priv_validator_key.json").open(
                "w"
            ) as fp:
                json.dump(
                    {
                        "address": hashlib.sha256(
                            base64.b64decode(val["consensus_key"]["pub"])
                        )
                        .hexdigest()[:40]
                        .upper(),
                        "pub_key": {
                            "type": "tendermint/PubKeyEd25519",
                            "value": val["consensus_key"]["pub"],
                        },
                        "priv_key": {
                            "type": "tendermint/PrivKeyEd25519",
                            "value": val["consensus_key"]["priv"],
                        },
                    },
                    fp,
                )
    if "genesis_file" in config:
        genesis_bytes = open(
            config["genesis_file"] % {"here": Path(config["path"]).parent}, "rb"
        ).read()
    else:
        genesis_bytes = (data_dir / "node0/config/genesis.json").read_bytes()
    (data_dir / "genesis.json").write_bytes(genesis_bytes)
    (data_dir / "gentx").mkdir()
    for i in range(len(config["validators"])):
        try:
            (data_dir / f"node{i}/config/genesis.json").unlink()
        except OSError:
            pass
        (data_dir / f"node{i}/config/genesis.json").symlink_to("../../genesis.json")
        (data_dir / f"node{i}/config/gentx").symlink_to("../../gentx")

    # now we can create ClusterCLI
    cli = ClusterCLI(data_dir.parent, chain_id=config["chain_id"], cmd=cmd)

    # patch the genesis file
    genesis = jsonmerge.merge(
        json.load(open(data_dir / "genesis.json")),
        config.get("genesis", {}),
    )
    (data_dir / "genesis.json").write_text(json.dumps(genesis))
    cli.validate_genesis()

    # create accounts
    accounts = []
    for i, node in enumerate(config["validators"]):
        mnemonic = node.get("mnemonic")
        account = cli.create_account("validator", i, mnemonic=mnemonic)
        accounts.append(account)
        if "coins" in node:
            cli.add_genesis_account(account["address"], node["coins"], i)
        if "staked" in node:
            cli.gentx(
                "validator",
                node["staked"],
                i=i,
                min_self_delegation=node.get("min_self_delegation", 1),
                pubkey=node.get("pubkey"),
            )

    # create accounts
    for account in config.get("accounts", []):
        account = create_account(cli, account)
        accounts.append(account)

    account_hw = config.get("hw_account")
    if account_hw:
        account = create_account(cli, account_hw, True)
        accounts.append(account)

    # output accounts
    (data_dir / "accounts.json").write_text(json.dumps(accounts))

    # collect-gentxs if directory not empty
    if next((data_dir / "gentx").iterdir(), None) is not None:
        cli.collect_gentxs(data_dir / "gentx", 0)

    # realise the symbolic links, so the node directories can be used independently
    genesis_bytes = (data_dir / "genesis.json").read_bytes()
    for i in range(len(config["validators"])):
        (data_dir / f"node{i}/config/gentx").unlink()
        tmp = data_dir / f"node{i}/config/genesis.json"
        tmp.unlink()
        tmp.write_bytes(genesis_bytes)

    # write tendermint config
    peers = config.get("peers") or ",".join(
        [
            "tcp://%s@%s:%d"
            % (cli.node_id(i), val["hostname"], ports.p2p_port(val["base_port"]))
            for i, val in enumerate(config["validators"])
        ]
    )
    for i, val in enumerate(config["validators"]):
        edit_tm_cfg(data_dir / f"node{i}/config/config.toml", val["base_port"], peers)
        edit_app_cfg(
            data_dir / f"node{i}/config/app.toml",
            val["base_port"],
            val.get("minimum-gas-prices", ""),
        )

    # write supervisord config file
    with (data_dir / SUPERVISOR_CONFIG_FILE).open("w") as fp:
        write_ini(fp, supervisord_ini(cmd, config["validators"], config["chain_id"]))

    if gen_compose_file:
        yaml.dump(
            docker_compose_yml(cmd, config["validators"], data_dir, image),
            (data_dir / "docker-compose.yml").open("w"),
        )


def relayer_chain_config(data_dir, chain):
    cfg = json.load((data_dir / chain["chain_id"] / "config.json").open())
    rpc_port = ports.rpc_port(cfg["validators"][0]["base_port"])
    grpc_port = ports.grpc_port(cfg["validators"][0]["base_port"])
    return {
        "key_name": "relayer",
        "id": chain["chain_id"],
        "rpc_addr": f"http://localhost:{rpc_port}",
        "grpc_addr": f"http://localhost:{grpc_port}",
        "websocket_addr": f"ws://localhost:{rpc_port}/websocket",
        "rpc_timeout": "10s",
        "account_prefix": chain.get("account-prefix", "cro"),
        "store_prefix": "ibc",
        "gas": 300000,
        "fee_denom": "basecro",
        "fee_amount": 0,
        "trusting_period": "336h",
    }


def init_cluster(
    data_dir, config_path, base_port, image=IMAGE, cmd=None, gen_compose_file=False
):
    config = yaml.safe_load(open(config_path))

    # override relayer config in config.yaml
    rly_section = config.pop("relayer", {})
    for chain_id, cfg in config.items():
        cfg["path"] = str(config_path)
        cfg["chain_id"] = chain_id

    chains = list(config.values())
    for chain in chains:
        (data_dir / chain["chain_id"]).mkdir()
        init_devnet(
            data_dir / chain["chain_id"], chain, base_port, image, cmd, gen_compose_file
        )
    paths = rly_section.get("paths", {})
    with (data_dir / SUPERVISOR_CONFIG_FILE).open("w") as fp:
        write_ini(
            fp,
            supervisord_ini_group(config.keys(), paths),
        )
    if len(chains) > 1:
        relayer_config = data_dir / "relayer.toml"
        # write relayer config
        relayer_config.write_text(
            tomlkit.dumps(
                {
                    "global": {
                        "strategy": "naive",
                        "log_level": "info",
                    },
                    "chains": [
                        relayer_chain_config(data_dir, chain) for chain in chains
                    ],
                    "connections": [
                        {
                            "a_chain": path["src"]["chain-id"],
                            "b_chain": path["dst"]["chain-id"],
                            "paths": [
                                {
                                    "a_port": path["src"]["port-id"],
                                    "b_port": path["dst"]["port-id"],
                                }
                            ],
                        }
                        for path in paths.values()
                    ],
                }
            )
        )

        # restore the relayer account in relayer
        for chain in chains:
            mnemonic = find_account(data_dir, chain["chain_id"], "relayer")["mnemonic"]
            subprocess.run(
                [
                    "hermes",
                    "-c",
                    relayer_config,
                    "keys",
                    "restore",
                    chain["chain_id"],
                    "--mnemonic",
                    mnemonic,
                    "--coin-type",
                    str(chain.get("coin-type", 394)),
                ],
                check=True,
            )


def find_account(data_dir, chain_id, name):
    accounts = json.load((data_dir / chain_id / "accounts.json").open())
    return next(acct for acct in accounts if acct["name"] == name)


def supervisord_ini(cmd, validators, chain_id):
    ini = {}
    for i, node in enumerate(validators):
        ini[f"program:{chain_id}-node{i}"] = dict(
            COMMON_PROG_OPTIONS,
            command=f"{cmd} start --home %(here)s/node{i}",
            stdout_logfile=f"%(here)s/node{i}.log",
        )
    return ini


def supervisord_ini_group(chain_ids, paths):
    cfg = {
        "include": {
            "files": " ".join(
                f"%(here)s/{chain_id}/tasks.ini" for chain_id in chain_ids
            )
        },
        "supervisord": {
            "pidfile": "%(here)s/supervisord.pid",
            "nodaemon": "true",
            "logfile": "/dev/null",
            "logfile_maxbytes": "0",
        },
        "rpcinterface:supervisor": {
            "supervisor.rpcinterface_factory": "supervisor.rpcinterface:"
            "make_main_rpcinterface",
        },
        "unix_http_server": {"file": "%(here)s/supervisor.sock"},
        "supervisorctl": {"serverurl": "unix://%(here)s/supervisor.sock"},
    }
    for path, path_cfg in paths.items():
        src = path_cfg["src"]["chain-id"]
        dst = path_cfg["dst"]["chain-id"]
        cfg[f"program:relayer-{path}"] = dict(
            COMMON_PROG_OPTIONS,
            command=(
                f"hermes -c %(here)s/relayer.toml start {src} {dst} "
                "-p transfer -c channel-0"
            ),
            stdout_logfile=f"%(here)s/relayer-{path}.log",
            autostart="false",
        )
    return cfg


def docker_compose_yml(cmd, validators, data_dir, image):
    return {
        "version": "3",
        "services": {
            f"node{i}": {
                "image": image,
                "command": "chaind start",
                "volumes": [f"{data_dir.absolute() / f'node{i}'}:/.chain-maind:Z"],
            }
            for i, val in enumerate(validators)
        },
    }


def edit_tm_cfg(path, base_port, peers, *, custom_edit=None):
    doc = tomlkit.parse(open(path).read())
    # tendermint is start in process, not needed
    # doc['proxy_app'] = 'tcp://127.0.0.1:%d' % abci_port(base_port)
    doc["rpc"]["laddr"] = "tcp://0.0.0.0:%d" % ports.rpc_port(base_port)
    doc["rpc"]["pprof_laddr"] = "localhost:%d" % ports.pprof_port(base_port)
    doc["rpc"]["grpc_laddr"] = "tcp://0.0.0.0:%d" % ports.grpc_port_tx_only(base_port)
    doc["p2p"]["laddr"] = "tcp://0.0.0.0:%d" % ports.p2p_port(base_port)
    doc["p2p"]["persistent_peers"] = peers
    doc["p2p"]["addr_book_strict"] = False
    doc["p2p"]["allow_duplicate_ip"] = True
    doc["consensus"]["timeout_commit"] = "1s"
    doc["rpc"]["timeout_broadcast_tx_commit"] = "30s"
    if custom_edit is not None:
        custom_edit(doc)
    open(path, "w").write(tomlkit.dumps(doc))


def edit_app_cfg(path, base_port, minimum_gas_prices=""):
    doc = tomlkit.parse(open(path).read())
    # enable api server
    doc["api"]["enable"] = True
    doc["api"]["swagger"] = True
    doc["api"]["enabled-unsafe-cors"] = True
    doc["api"]["address"] = "tcp://0.0.0.0:%d" % ports.api_port(base_port)
    doc["grpc"]["address"] = "0.0.0.0:%d" % ports.grpc_port(base_port)
    # take snapshot for statesync
    doc["pruning"] = "nothing"
    doc["state-sync"]["snapshot-interval"] = 5
    doc["state-sync"]["snapshot-keep-recent"] = 10
    doc["minimum-gas-prices"] = minimum_gas_prices
    open(path, "w").write(tomlkit.dumps(doc))


if __name__ == "__main__":
    interact("rm -r data; mkdir data", ignore_error=True)
    data_dir = Path("data")
    init_cluster(data_dir, "config.yaml", 26650)
    supervisord = start_cluster(data_dir)
    t = start_tail_logs_thread(data_dir)
    supervisord.wait()
    t.stop()
    t.join()



import os

CHAIN = ""  # edit by nix-build
if not CHAIN:
    CHAIN = os.environ.get("CHAIN_MAIND", "chain-maind")
IMAGE = "docker.pkg.github.com/crypto-org-chain/chain-main/chain-main-pystarport:latest"

SUPERVISOR_CONFIG_FILE = "tasks.ini"



import random
import sys
import threading
import time

import yaml

from .cluster import ClusterCLI
from .cosmoscli import CosmosCLI


class TxJobThread(threading.Thread):
    def __init__(self, label, job, cosmos_cli: CosmosCLI):
        threading.Thread.__init__(self)
        self.label = label
        self.job = job
        self.cosmos_cli = cosmos_cli

    def transfer_tx_job(self):
        from_address = self.cosmos_cli.address(
            self.job["from_account"],
        )
        to_address = self.job["to_address"]
        if "random_amount" in self.job:
            amount = random_amount(
                self.job["random_amount"][0],
                self.job["random_amount"][1],
                self.job["random_amount"][2],
            )
        else:
            amount = self.job["amount"]

        print(
            "[%s] Transfer %s from %s to %s"
            % (self.label, amount, from_address, to_address)
        )
        result = self.cosmos_cli.transfer(from_address, to_address, amount)
        print(result)

    def delegate_tx_job(self):
        from_address = self.cosmos_cli.address(self.job["from_account"])
        to_address = self.job["to_validator_address"]
        if "random_amount" in self.job:
            amount = random_amount(
                self.job["random_amount"][0],
                self.job["random_amount"][1],
                self.job["random_amount"][2],
            )
        else:
            amount = self.job["amount"]

        print(
            "[%s] Delegate %s from %s to %s"
            % (self.label, amount, from_address, to_address)
        )
        result = self.cosmos_cli.delegate_amount(to_address, amount, from_address)
        print(result)

    def withdraw_all_rewards_job(self):
        from_address = self.cosmos_cli.address(self.job["from_account"])
        print("[%s] Withdraw all rewards from %s" % (self.label, from_address))
        result = self.cosmos_cli.withdraw_all_rewards(from_address)
        print(result)

    def next_interval(self):
        if "random_interval" in self.job:
            return random.randint(
                self.job["random_interval"][0],
                self.job["random_interval"][1],
            )
        return self.job["interval"]

    def run(self):
        job_type = self.job["type"]
        while 1:
            begin = time.time()

            try:
                if job_type == "transfer":
                    self.transfer_tx_job()
                elif job_type == "delegate":
                    self.delegate_tx_job()
                elif job_type == "withdraw_all_rewards":
                    self.withdraw_all_rewards_job()
                else:
                    print("Unknown job type: %s", job_type)
                    sys.exit()
            except Exception as e:
                print("error executing job:", sys.exc_info(), str(e))

            interval = self.next_interval()

            duration = time.time() - begin
            if duration < interval:
                sleep = interval - duration
                print("Next %s in %ds ...\n" % (job_type, sleep))
                time.sleep(sleep)


def random_amount(min, max, denom):
    return "%d%s" % (random.randint(min, max), denom)


class BotClusterCLI:
    "transaction bot Cluster CLI"

    def __init__(self, config_path, cluster_cli: ClusterCLI):
        self.config = yaml.safe_load(open(config_path))
        self.cluster_cli = cluster_cli

    def start(self):
        """
        prepare and start a transaction bot from configuration
        """
        threads = []
        for i, job in enumerate(self.config["jobs"], start=1):
            node_i = job.get("node", 0)
            cli = CosmosCLI(
                self.cluster_cli.home(node_i),
                self.cluster_cli.node_rpc(node_i),
                chain_id=self.cluster_cli.chain_id,
                cmd=self.cluster_cli.cmd,
            )
            thread = TxJobThread(job.get("label", i), job, cli)

            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()


class BotCLI:
    "transaction bot CLI"

    def __init__(self, config_path, cosmos_cli=None):
        self.config = yaml.safe_load(open(config_path))
        self.cosmos_cli = cosmos_cli

    def start(self):
        """
        prepare and start a transaction bot from configuration
        """
        threads = []
        for i, job in enumerate(self.config["jobs"], start=1):
            thread = TxJobThread(job.get("label", i), job, self.cosmos_cli)

            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()



import requests


class ApiUtil:
    def __init__(self, ip_port):
        self.base_url = f"http://127.0.0.1:{ip_port}"

    def balance(self, address):
        url = f"{self.base_url}/cosmos/bank/v1beta1/balances/{address}"
        response = requests.get(url)
        balance = response.json()["balances"]
        if len(balance) > 0:
            return int(balance[0]["amount"])
        else:
            return 0

    def account_info(self, address):
        url = f"{self.base_url}/cosmos/auth/v1beta1/accounts/{address}"
        response = requests.get(url)
        account_info = response.json()["account"]
        account_num = int(account_info["account_number"])
        sequence = int(account_info["sequence"])
        return {"account_num": account_num, "sequence": sequence}

    def broadcast_tx(self, signed_tx: dict):
        url = f"{self.base_url}/txs"
        response = requests.post(url, json=signed_tx)
        if not response.ok:
            raise Exception(
                f"response code: {response.status_code}, {response.reason}, {response.json()}"
            )
        result = response.json()
        if result.get("code"):
            raise Exception(result["raw_log"])
        return result



import configparser
import subprocess


def interact(cmd, ignore_error=False, input=None, **kwargs):
    proc = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        **kwargs,
    )
    # begin = time.perf_counter()
    (stdout, _) = proc.communicate(input=input)
    # print('[%.02f] %s' % (time.perf_counter() - begin, cmd))
    if not ignore_error:
        assert proc.returncode == 0, f'{stdout.decode("utf-8")} ({cmd})'
    return stdout


def write_ini(fp, cfg):
    ini = configparser.RawConfigParser()
    for section, items in cfg.items():
        ini.add_section(section)
        sec = ini[section]
        sec.update(items)
    ini.write(fp)


def safe_cli_string(s):
    'wrap string in "", used for cli argument when contains spaces'
    if len(f"{s}".split()) > 1:
        return f'"{s}"'
    return f"{s}"


def build_cli_args_safe(*args, **kwargs):
    args = [safe_cli_string(arg) for arg in args if arg]
    for k, v in kwargs.items():
        if v is None:
            continue
        args.append("--" + k.strip("_").replace("_", "-"))
        args.append(safe_cli_string(v))
    return list(map(str, args))


def build_cli_args(*args, **kwargs):
    args = [arg for arg in args if arg is not None]
    for k, v in kwargs.items():
        if v is None:
            continue
        args.append("--" + k.strip("_").replace("_", "-"))
        args.append(v)
    return list(map(str, args))


def format_doc_string(**kwargs):
    def decorator(target):
        target.__doc__ = target.__doc__.format(**kwargs)
        return target

    return decorator



import socket
import time
import uuid

import docker

ZEMU_HOST = "127.0.0.1"
ZEMU_BUTTON_PORT = 9997
ZEMU_GRPC_SERVER_PORT = 3002
# dockerfile is integration_test/hardware_wallet/Dockerfile
ZEMU_IMAGE = "cryptocom/builder-zemu:latest"


class Ledger:
    def __init__(self):
        self.ledger_name = f"ledger_simulator_{uuid.uuid4().time_mid}"
        self.proxy_name = f"ledger_proxy_{uuid.uuid4().time_mid}"
        self.grpc_name = f"ledger_grpc_server_{uuid.uuid4().time_mid}"
        self.cmds = {
            self.ledger_name: [
                "./speculos/speculos.py",
                "--display=headless",
                f"--button-port={ZEMU_BUTTON_PORT}",
                "./speculos/apps/crypto.elf",
            ],
            self.proxy_name: ["./speculos/tools/ledger-live-http-proxy.py", "-v"],
            self.grpc_name: ["bash", "-c", "RUST_LOG=debug zemu-grpc-server"],
        }
        self.client = docker.from_env()
        self.client.images.pull(ZEMU_IMAGE)
        self.containers = []

    def start(self):
        host_config_ledger = self.client.api.create_host_config(
            auto_remove=True,
            port_bindings={
                ZEMU_BUTTON_PORT: ZEMU_BUTTON_PORT,
                ZEMU_GRPC_SERVER_PORT: ZEMU_GRPC_SERVER_PORT,
            },
        )
        container_ledger = self.client.api.create_container(
            ZEMU_IMAGE,
            self.cmds[self.ledger_name],
            name=self.ledger_name,
            ports=[ZEMU_BUTTON_PORT, ZEMU_GRPC_SERVER_PORT],
            host_config=host_config_ledger,
        )
        self.client.api.start(container_ledger["Id"])
        self.containers.append(container_ledger)
        for name in [self.proxy_name, self.grpc_name]:
            cmd = self.cmds[name]
            try:
                host_config = self.client.api.create_host_config(
                    auto_remove=True, network_mode=f"container:{self.ledger_name}"
                )
                container = self.client.api.create_container(
                    ZEMU_IMAGE,
                    cmd,
                    name=name,
                    host_config=host_config,
                )
                self.client.api.start(container["Id"])
                self.containers.append(container)
                time.sleep(2)
            except Exception as e:
                print(e)

    def stop(self):
        for container in self.containers:
            try:
                self.client.api.remove_container(container["Id"], force=True)
                print("stop docker {}".format(container["Name"]))
            except Exception as e:
                print(e)


class LedgerButton:
    def __init__(self, zemu_address, zemu_button_port):
        self._client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.zemu_address = zemu_address
        self.zemu_button_port = zemu_button_port
        self.connected = False

    @property
    def client(self):
        if not self.connected:
            time.sleep(5)
            self._client.connect((self.zemu_address, self.zemu_button_port))
            self.connected = True
        return self._client

    def press_left(self):
        data = "Ll"
        self.client.send(data.encode())

    def press_right(self):
        data = "Rr"
        self.client.send(data.encode())

    def press_both(self):
        data = "LRlr"
        self.client.send(data.encode())



import os
import signal
from pathlib import Path

import fire

from .app import CHAIN, IMAGE, SUPERVISOR_CONFIG_FILE
from .bot import BotCLI, BotClusterCLI
from .cluster import ClusterCLI, init_cluster, start_cluster, start_tail_logs_thread
from .cosmoscli import CosmosCLI
from .utils import build_cli_args, interact


def init(
    data,
    config,
    base_port,
    *args,
    **kwargs,
):
    interact(
        f"rm -r {data}; mkdir {data}",
        ignore_error=True,
    )
    return init_cluster(data, config, base_port, *args, **kwargs)


def start(data, quiet):
    supervisord = start_cluster(data)

    # register signal to quit supervisord
    for signame in ("SIGINT", "SIGTERM"):
        signal.signal(getattr(signal, signame), lambda *args: supervisord.terminate())

    if not quiet:
        tailer = start_tail_logs_thread(data)

    supervisord.wait()

    if not quiet:
        tailer.stop()
        tailer.join()


def serve(data, config, base_port, cmd, quiet):
    init(data, config, base_port, cmd=cmd)
    start(data, quiet)


class CLI:
    def __init__(self, /, cmd=CHAIN):
        """
        :param cmd: path to the chain binary
        """
        self.cmd = cmd

    def init(
        self,
        data: str = "./data",
        config: str = "./config.yaml",
        base_port: int = 26650,
        image: str = IMAGE,
        gen_compose_file: bool = False,
        cmd: str = CHAIN,
    ):
        """
        prepare all the configurations of a devnet

        :param data: path to the root data directory
        :param config: path to the configuration file
        :param base_port: the base port to use, the service ports of different nodes
        are calculated based on this
        :param image: the image used in the generated docker-compose.yml
        :param gen_compose_file: generate a docker-compose.yml
        :param cmd: path to chain binary
        """
        init(Path(data), config, base_port, image, self.cmd, gen_compose_file)

    def start(self, data: str = "./data", quiet: bool = False):
        """
        start the prepared devnet

        :param data: path to the root data directory
        :param quiet: don't print logs of subprocesses
        """
        start(Path(data), quiet)

    def chaind(self, *args, **kwargs):
        """
        start one node whose home directory is already initialized
        can be used to launch chain-maind

        :param home: home directory
        """
        os.execvp(self.cmd, [self.cmd] + build_cli_args(*args, **kwargs))

    def serve(
        self,
        data: str = "./data",
        config: str = "./config.yaml",
        base_port: int = 26650,
        quiet: bool = False,
    ):
        """
        prepare and start a devnet from scatch

        :param data: path to the root data directory
        :param config: path to the configuration file
        :param base_port: the base port to use, the service ports of different nodes
        are calculated based on this
        :param cmd: path to chain binary
        :param quiet: don't print logs of subprocesses
        """
        serve(Path(data), config, base_port, self.cmd, quiet)

    def supervisorctl(self, *args, data: str = "./data"):
        from supervisor.supervisorctl import main

        main(("-c", Path(data) / SUPERVISOR_CONFIG_FILE, *args))

    def cli(self, *args, data: str = "./data", chain_id: str = "chainmaind"):
        """
        pystarport CLI

        :param data: path to the root data directory
        :param chain_id: chain id of the cluster
        """
        return ClusterCLI(Path(data), chain_id=chain_id, cmd=self.cmd)

    def bot(
        self,
        *args,
        data: str = "./data",
        config: str = "./bot.yaml",
        chain_id: str = "chainmaind",
        node_rpc: str = None,
    ):
        """
        transaction bot CLI

        :param data: path to the root data directory if connecting to pystarport
        cluster. Path to the home directory if connecting to a node
        :param config: path to the bot configuration file
        (copy bot.yaml.example for reference)
        :param chain_id: chain id of the cluster
        :param node_rpc: custom Tendermint RPC endpoint to the node
        """
        data_path = Path(data)
        config_path = Path(config)
        if node_rpc is None:
            cluster_cli = ClusterCLI(data_path, chain_id=chain_id, cmd=self.cmd)
            return BotClusterCLI(config_path, cluster_cli)
        else:
            cosmos_cli = CosmosCLI(data_path, node_rpc, cmd=self.cmd)
            return BotCLI(config_path, cosmos_cli)


def main():
    fire.Fire(CLI)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pystarport', 'pystarport.proto_python']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'bech32>=1.1.0,<2.0.0',
 'docker>=4.3.1,<5.0.0',
 'durations>=0.3.3,<0.4.0',
 'fire>=0.4.0,<0.5.0',
 'jsonmerge>=1.7.0,<2.0.0',
 'multitail2>=1.5.2,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'supervisor>=4.2.1,<5.0.0',
 'tomlkit>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['pystarport = pystarport.cli:main']}

setup_kwargs = {
    'name': 'pystarport',
    'version': '0.1.0',
    'description': 'Spawn local devnet for cosmos-sdk chains',
    'long_description': 'pystarport is like a [cosmos starport](https://github.com/tendermint/starport)\nwithout the scaffolding feature. it\'s mainly used for development and testing. It\'s developed for the Crypto.org Chain, but\nit can also be used for any cosmos-sdk based projects.\n\n## Configuration\n\na typical configuration for a devnet is like this:\n\n```\nchainmaind:\n  cmd: chain-maind  # chain binary to use, optional\n  validators:  # genesis validators\n    - coins: 10cro\n      staked: 10cro\n    - coins: 10cro\n      staked: 10cro\n  accounts:  # genesis accounts\n    - name: community\n      coins: 100cro\n    - name: ecosystem\n      coins: 200cro\n    - name: reserve\n      coins: 200cro\n      vesting: "1d"\n    - name: launch\n      coins: 100cro\n  genesis:  # patch genesis states\n   app_state:\n     staking:\n       params:\n         unbonding_time: "10s"\n```\n\nThe `validators` section defines how many nodes to run, for each node, a home directory is initialized in\n`data/node{i}`, and a validator account with specified coins is created.\n\nThe `accounts` defines other non-validator accounts, they are created in `node0`\'s keyring.\n\nIn the `genesis` section you can override any genesis configuration with the same json path.\n\n## Usage\n\n```\nNAME\n    pystarport serve - prepare and start a devnet from scatch\n\nSYNOPSIS\n    pystarport serve <flags>\n\nDESCRIPTION\n    prepare and start a devnet from scatch\n\nFLAGS\n    --data=DATA\n        Type: str\n        Default: \'./data\'\n        path to the root data directory\n    --config=CONFIG\n        Type: str\n        Default: \'./config.yaml\'\n        path to the configuration file\n    --base_port=BASE_PORT\n        Type: int\n        Default: 26650\n        the base port to use, the service ports of different nodes are calculated based on this\n    --cmd=CMD\n        Type: str\n        Default: \'chain-maind\'\n        the chain binary to use\n```\n\n## Port rules\n\nThe rules to calculate service ports based on base port is defined in the\n[`ports.py`](https://github.com/crypto-org-chain/chain-main/blob/master/pystarport/pystarport/ports.py) module.\n\nFor example, with default base port `26650`, the url of api servers of the nodes would be:\n\n- Node0: http://127.0.0.1:26654\n- Node1: http://127.0.0.1:26664\n\n> The swagger doc of node0 is http://127.0.0.1:26654/swagger/\n>\n> The default rpc port used by `chain-maind` is `26657`, that\'s the default node0\'s rpc port, so you can use\n> `chain-maind` without change to access node0\'s rpc.\n\n## Supervisor\n\n`pystarport` embeded a [supervisor](http://supervisord.org/) to manage processes of multiple nodes, you can use\n`pystarport supervisorctl` to manage the processes:\n\n```\n$ pystarport supervisorctl status\nnode0                            RUNNING   pid 35210, uptime 0:00:29\nnode1                            RUNNING   pid 35211, uptime 0:00:29\n$ pystarport supervisorctl help\n\ndefault commands (type help <topic>):\n=====================================\nadd    exit      open  reload  restart   start   tail\navail  fg        pid   remove  shutdown  status  update\nclear  maintail  quit  reread  signal    stop    version\n```\n\nOr enter an interactive shell:\n\n```\n$ pystarport supervisorctl\nnode0                            RUNNING   pid 35210, uptime 0:01:53\nnode1                            RUNNING   pid 35211, uptime 0:01:53\nsupervisor>\n```\n\n## Cli\n\nAfter started the chain, you can use `chain-maind` cli directly, there are also some wrapper commands provided by\n`pystarport cli`. It understands the directory structure and port rules, also assuming `keyring-backend=test`, and there\nare shortcuts for commonly used commands, so arguments are shorter.\n\n```\n$ pystarport cli - --help\n...\n```\n\n## Transaction Bot\n\nA simple transaction bot that works for cluster created by pystarport as well as a local node\n\nCopy and modify `bot.yaml.sample` to `bot.yaml` with your desired bot configurations.\n\n### If you are running on a pystarport created cluster:\n1. Make sure you have provide the `node` for each job in the `bot.yaml`\n2. Run the command\n```\n$ pystarport bot --chain-id=[cluster_chain_id] - start\n```\n\n### If you are running on a local node\n```\n$ pstarport bot --node_rpc=tcp://127.0.0.1:26657 --data=/path/to/your/local/node/home/ - start\n```\n\n## docker-compose\n\nWhen used with `docker-compose` or multiple machines, you need to config hostnames, and you probabely want to use a same\n`base_port` since you don\'t have port conflicts, you can config the `validators` like this:\n\n```yaml\nvalidators:\n  - coins: 10cro\n    staked: 10cro\n    base_port: 26650\n    hostname: node0\n  - coins: 10cro\n    staked: 10cro\n    base_port: 26650\n    hostname: node1\n```\n\n`pystarport init --gen_compose_file` will also generate a `docker-compose.yml` file for you.\n\n## IBC\n\nIt can setup multiple devnets at once, and connect them with ibc relayer.\n\n```\nibc-0:\n  validators:\n    - coins: 10cro\n      staked: 10cro\n      base_port: 26650\n    - coins: 10cro\n      staked: 10cro\n  accounts:\n    - name: relayer\n      coins: 100cro\n  genesis:\n    app_state:\n      transfer:\n        params:\n          receive_enabled: true\n          send_enabled: true\nibc-1:\n  validators:\n    - coins: 10cro\n      staked: 10cro\n      base_port: 26750\n    - coins: 10cro\n      staked: 10cro\n      base_port: 26760\n  accounts:\n    - name: relayer\n      coins: 100cro\n  genesis:\n    app_state:\n      transfer:\n        params:\n          receive_enabled: true\n          send_enabled: true\nrelayer:\n  global:\n    timeout: 10s\n    light-cache-size: 20\n  paths:\n    demo:\n      src:\n        chain-id: ibc-0\n        port-id: transfer\n        order: unordered\n        version: ics20-1\n      dst:\n        chain-id: ibc-1\n        port-id: transfer\n        order: unordered\n        version: ics20-1\n      strategy:\n        type: naive\n```\n\nWith following commands to setup ibc, you are ready to play with ibc functionalities:\n\n```\n# spawn the devnets\npystarport serve --config ibc.yaml\n# setup ibc channel\nhermes -c data/relayer.toml create channel ibc-0 ibc-1 --port-a transfer --port-b transfer\n# start relayer process\nsupervisorctl -c data/tasks.ini relayer-demo\n```\n\n',
    'author': 'huangyi',
    'author_email': 'huang@crypto.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/crypto-org-chain/chain-main/tree/master/pystarport',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
